@@ -113,7 +113,6 @@ md5sum (const char *filename, char *result)
   result[32] = '\\0';
 }
 
-#if 0 /* <- Remove this if we add RHashtable tests in 1.14 branch. */
 /* Return the value for a key in a hashtable.
  * Note: the return value is part of the hash and should not be freed.
  */
@@ -129,7 +128,6 @@ get_key (char **hash, const char *key)
 
   return NULL; /* key not found */
 }
-#endif
 
 ";
 
